// full_topology.cpp

#include <chrono>
#include <memory>
#include <random>
#include <string>
#include <thread>
#include <vector>
// the order here matters
#include "tracetools/utils.hpp"

#include <rclcpp/rclcpp.hpp>
#include "std_msgs/msg/string.hpp"

/** Whether to use timers in fusion nodes */
//#define USE_TIMER_IN_FUSION_NODES

using namespace std::chrono_literals;

constexpr bool DEBUG = false;

// A generic sensor node that publishes at a given rate + jitter
#define DEFINE_SENSOR_NODE_CLASS(NAME)                                   \
class NAME : public rclcpp::Node                                         \
{                                                                        \
public:                                                                  \
  NAME(std::string const &name,                                          \
             std::string const &topic,                                   \
             int rate_hz,                                                \
             double jitter_ms_ = 0.0)                                    \
      : Node(name)                                                       \
  {                                                                      \
    publisher_ = create_publisher<std_msgs::msg::String>(topic, 10);     \
    auto period = std::chrono::milliseconds(1000 / rate_hz);             \
    timer_ = create_wall_timer(period, [this, name, jitter_ms_]()        \
                               {                                         \
      double jitter = 0.0;                                               \
      if (jitter_ms_ > 0.0) {                                            \
        static thread_local std::mt19937 gen(std::random_device{}());    \
        std::normal_distribution<> d(0.0, jitter_ms_);                   \
        jitter = d(gen);                                                 \
      }                                                                  \
      /* Ensure total delay is non-negative */                           \
      int total_delay = std::max(0, static_cast<int>(std::round(jitter))); \
      std::this_thread::sleep_for(std::chrono::milliseconds(total_delay)); \
      auto msg = std_msgs::msg::String();                                \
      msg.data = name + std::string(" publishes");                       \
      if (DEBUG) {                                                       \
        RCLCPP_INFO(get_logger(), "%s", msg.data.c_str());               \
      }                                                                  \
      publisher_->publish(msg); });                                      \
  }                                                                      \
                                                                         \
private:                                                                 \
  rclcpp::Publisher<std_msgs::msg::String>::SharedPtr publisher_;        \
  rclcpp::TimerBase::SharedPtr timer_;                                   \
};

// A generic single-input processing node with a fixed delay +/- jitter
#define DEFINE_PROCESSING_NODE_CLASS(NAME)                               \
class NAME : public rclcpp::Node                                         \
{                                                                        \
public:                                                                  \
  NAME(const std::string & name,                                         \
       const std::string & input_topic,                                  \
       const std::string & output_topic,                                 \
       int delay_ms,                                                     \
       double jitter_ms = 0.0)                                           \
    : Node(name), delay_ms_(delay_ms), jitter_ms_(jitter_ms)             \
  {                                                                      \
    subscription_ = create_subscription<std_msgs::msg::String>(          \
        input_topic, 10,                                                 \
        [this, name](std_msgs::msg::String::SharedPtr msg)               \
        {                                                                \
          if (DEBUG) {                                                   \
            RCLCPP_INFO(get_logger(), "%s received: %s", name.c_str(), msg->data.c_str()); \
          }                                                              \
          /* Simulate processing delay */                                \
          double jitter = 0.0;                                           \
          if (jitter_ms_ > 0.0) {                                        \
            static thread_local std::mt19937 gen(std::random_device{}());\
            std::normal_distribution<> d(0.0, jitter_ms_);               \
            jitter = d(gen);                                             \
          }                                                              \
          std::this_thread::sleep_for(std::chrono::milliseconds((int)(delay_ms_ + jitter))); \
          auto out = std_msgs::msg::String();                            \
          out.data = msg->data + " -> " + name;                          \
          publisher_->publish(out);                                      \
        });                                                              \
    publisher_ = create_publisher<std_msgs::msg::String>(output_topic, 10); \
  }                                                                      \
protected:                                                               \
  int delay_ms_;                                                         \
  double jitter_ms_;                                                     \
  rclcpp::Subscription<std_msgs::msg::String>::SharedPtr subscription_;  \
  rclcpp::Publisher<std_msgs::msg::String>::SharedPtr publisher_;        \
};

// A generic single-input processing node with a fixed delay +/- jitter
#define DEFINE_MULTI_PROCESSING_NODE_CLASS(NAME)                         \
class NAME : public rclcpp::Node                                         \
{                                                                        \
public:                                                                  \
  NAME(std::string const &name,                                          \
        std::vector<std::string> const &input_topics,                    \
        std::string const &output_topic,                                 \
        int delay_ms,                                                    \
        double jitter_ms = 0.0)                                          \
    : Node(name), delay_ms_(delay_ms), jitter_ms_(jitter_ms)             \
  {                                                                      \
    for (auto &topic : input_topics)                                     \
    {                                                                    \
      auto sub = create_subscription<std_msgs::msg::String>(             \
          topic, 10,                                                     \
          [this, name](std_msgs::msg::String::SharedPtr msg)             \
          {                                                              \
            /* simple fusion: log each arrival */                        \
            if (DEBUG)                                                   \
            {                                                            \
              RCLCPP_INFO(get_logger(), "%s fusing input: %s", name.c_str(), msg->data.c_str()); \
            }                                                            \
            /* Simulate processing delay */                              \
            double jitter = 0.0;                                         \
            if (jitter_ms_ > 0.0) {                                      \
              static thread_local std::mt19937 gen(std::random_device{}()); \
              std::normal_distribution<> d(0.0, jitter_ms_);             \
              jitter = d(gen);                                           \
            }                                                            \
            std::this_thread::sleep_for(std::chrono::milliseconds((int)(delay_ms_ + jitter))); \
            auto out = std_msgs::msg::String();                          \
            out.data = msg->data + " -> " + name;                        \
            if (DEBUG) {                                                 \
              RCLCPP_INFO(get_logger(), "%s publishes fused message", name.c_str()); \
            }                                                            \
            publisher_->publish(out);                                    \
          });                                                            \
      subscriptions_.emplace_back(sub);                                  \
    }                                                                    \
    publisher_ = create_publisher<std_msgs::msg::String>(output_topic, 10); \
                                                                         \
  }                                                                      \
protected:                                                               \
  std::vector<rclcpp::Subscription<std_msgs::msg::String>::SharedPtr> subscriptions_; \
  rclcpp::Publisher<std_msgs::msg::String>::SharedPtr publisher_;        \
  int delay_ms_;                                                         \
  double jitter_ms_;                                                     \
};

DEFINE_SENSOR_NODE_CLASS(CameraNode)
DEFINE_PROCESSING_NODE_CLASS(DebayerNode)
DEFINE_PROCESSING_NODE_CLASS(RadiometricNode)
DEFINE_PROCESSING_NODE_CLASS(GeometricNode)
DEFINE_PROCESSING_NODE_CLASS(MappingNode)

DEFINE_PROCESSING_NODE_CLASS(SmokeClassifierNode)

DEFINE_SENSOR_NODE_CLASS(GPSNode)
DEFINE_SENSOR_NODE_CLASS(IMUNode)
DEFINE_SENSOR_NODE_CLASS(BaroNode)

DEFINE_PROCESSING_NODE_CLASS(FusionNodeS)
DEFINE_MULTI_PROCESSING_NODE_CLASS(FusionNode)

DEFINE_SENSOR_NODE_CLASS(LidarNode)

DEFINE_SENSOR_NODE_CLASS(CommandNode)

DEFINE_PROCESSING_NODE_CLASS(FlightManagementNodeS)

DEFINE_MULTI_PROCESSING_NODE_CLASS(FlightManagementNode)

DEFINE_PROCESSING_NODE_CLASS(ControlNode)

DEFINE_PROCESSING_NODE_CLASS(TelemetryNodeS)
DEFINE_MULTI_PROCESSING_NODE_CLASS(TelemetryNode)

DEFINE_PROCESSING_NODE_CLASS(RadioNode)

#define PROCENTUAL_JITTER 0.1
#define DELAY_JITTER_PAIR(DELAY) DELAY, PROCENTUAL_JITTER * DELAY

int main(int argc, char **argv)
{
  rclcpp::init(argc, argv);
  rclcpp::executors::MultiThreadedExecutor executor;

  // Chain 1: Image2Ground Mapping
  auto camera      = std::make_shared<CameraNode>("input_camera_node", "/input/camera/raw", DELAY_JITTER_PAIR(5));     // Sensor read, low latency
  auto debayer     = std::make_shared<DebayerNode>("debayer_node", "/input/camera/raw", "/camera/debayered", DELAY_JITTER_PAIR(18));  // Moderate, image conversion
  auto radiometric = std::make_shared<RadiometricNode>("radiometric_node", "/camera/debayered", "/camera/radiometric", DELAY_JITTER_PAIR(22)); // Heavier image math
  auto geometric   = std::make_shared<GeometricNode>("geometric_node", "/camera/radiometric", "/camera/geometric", DELAY_JITTER_PAIR(12));    // Perspective transform, medium
  auto mapping     = std::make_shared<MappingNode>("output_mapping_node", "/camera/geometric", "/output/camera/mapped", DELAY_JITTER_PAIR(16)); // Warping/stitching, moderate

  // Chain 2: Semantic Scheduler Adaptation
  auto smoke = std::make_shared<SmokeClassifierNode>("output_smoke_classifier_node", "/camera/radiometric", "/output/classifier/classification", DELAY_JITTER_PAIR(35)); // CNN/ML classifier, can be high

  // Chain 3: Flight Control
  auto gps    = std::make_shared<GPSNode>("input_gps_node", "/input/gps/fix", DELAY_JITTER_PAIR(7));    // Fast, low latency sensor
  auto imu    = std::make_shared<IMUNode>("input_imu_node", "/input/imu/data", DELAY_JITTER_PAIR(5));   // Fast, low latency sensor
  auto baro   = std::make_shared<BaroNode>("input_baro_node", "/input/baro/alt", DELAY_JITTER_PAIR(6)); // Fast, low latency sensor
  auto fusion = std::make_shared<FusionNode>(
      "sensor_fusion_node", std::vector<std::string>{"/input/gps/fix", "/input/imu/data", "/input/baro/alt"}, "/sensors/fused", DELAY_JITTER_PAIR(14)); // Sensor fusion, not trivial but not heavy

  auto lidar  = std::make_shared<LidarNode>("input_lidar_node", "/input/lidar/scan", DELAY_JITTER_PAIR(10)); // LIDAR driver, can be fast
  auto cmd    = std::make_shared<CommandNode>("input_operator_cmd_node", "/input/operator/commands", DELAY_JITTER_PAIR(8)); // Operator commands, fast
  auto mgmt   = std::make_shared<FlightManagementNode>(
      "flight_mgmt_node", std::vector<std::string>{"/sensors/fused", "/input/lidar/scan", "/input/operator/commands"}, "/flight/plan", DELAY_JITTER_PAIR(18)); // Path planning, moderate
  auto control = std::make_shared<ControlNode>("output_flight_control_node", "/flight/plan", "/output/flight/cmd", DELAY_JITTER_PAIR(12)); // PID controller, fast but non-zero

  // Chain 4: Data Storage/Streaming
  auto telemetry = std::make_shared<TelemetryNode>(
      "telemetry_node", std::vector<std::string>{"/sensors/fused", "/output/camera/mapped"}, "/telemetry/data", DELAY_JITTER_PAIR(6)); // Quick serialization
  auto radio = std::make_shared<RadioNode>("output_radio_tx_node", "/telemetry/data", "/output/telemetry/radio", DELAY_JITTER_PAIR(3)); // Output, almost instant

  // Add all to executor
  executor.add_node(camera);
  executor.add_node(debayer);
  executor.add_node(radiometric);
  executor.add_node(geometric);
  executor.add_node(mapping);
  
  executor.add_node(smoke);

  executor.add_node(gps);
  executor.add_node(imu);
  executor.add_node(baro);
  executor.add_node(fusion);
  
  executor.add_node(telemetry);
  executor.add_node(radio);

  executor.add_node(lidar);
  executor.add_node(cmd);
  executor.add_node(mgmt);
  executor.add_node(control);

  executor.spin();

  rclcpp::shutdown();
  return 0;
}
